@@ -8,7 +8,7 @@
 // Version 3.0 developed by Mihai Bazon for InteractiveTools.
 //	     http://students.infoiasi.ro/~mishoo
 //
-// $Id: htmlarea.js,v 1.18 2003/05/31 11:51:52 mishoo Exp $
+// $Id: htmlarea.js,v 1.19 2003/08/10 15:56:34 mishoo Exp $
 
 // Creates a new HTMLArea object.  Tries to replace the textarea with the given
 // ID with it.
@@ -22,6 +22,9 @@ function HTMLArea(textarea, config) {
 		this._htmlArea = null;
 		this._textArea = textarea;
 		this._editMode = "wysiwyg";
+		this.plugins = {};
+		this._timerToolbar = null;
+		this._mdoc = document; // cache the document, we need it in plugins
 	}
 };
 
@@ -31,6 +34,9 @@ HTMLArea.Config = function () {
 	this.width = "auto";
 	this.height = "auto";
 
+	// enable creation of a status bar?
+	this.statusBar = true;
+
 	// the next parameter specifies whether the toolbar should be included
 	// in the size or not.
 	this.sizeIncludesToolbar = true;
@@ -47,7 +53,8 @@ HTMLArea.Config = function () {
 	this.imgURL = "images/";
 	this.popupURL = "popups/";
 
-	this.debug = 0;
+	// configuration for plugins
+	this.plugins = {};
 
 	/** CUSTOMIZING THE TOOLBAR
 	 * -------------------------
@@ -58,16 +65,19 @@ HTMLArea.Config = function () {
 	 * new official version, it's less likely that you will have problems
 	 * upgrading HTMLArea.
 	 */
-	this.toolbar = [ [ "fontname", "space" ],
-			 [ "fontsize", "space" ],
-			 [ "formatblock", "space"],
-			 [ "bold", "italic", "underline", "separator" ],
-			 [ "strikethrough", "subscript", "superscript", "linebreak" ],
-			 [ "justifyleft", "justifycenter", "justifyright", "justifyfull", "separator" ],
-			 [ "insertorderedlist", "insertunorderedlist", "outdent", "indent", "separator" ],
-			 [ "forecolor", "hilitecolor", "textindicator", "separator" ],
-			 [ "inserthorizontalrule", "createlink", "insertimage", "inserttable", "htmlmode", "separator" ],
-			 [ "popupeditor", "separator", "showhelp", "about" ]
+	this.toolbar = [
+		[ "fontname", "space",
+		  "fontsize", "space",
+		  "formatblock", "space",
+		  "bold", "italic", "underline", "separator",
+		  "strikethrough", "subscript", "superscript", "separator",
+		  "copy", "cut", "paste", "space", "undo", "redo" ],
+		
+		[ "justifyleft", "justifycenter", "justifyright", "justifyfull", "separator",
+		  "insertorderedlist", "insertunorderedlist", "outdent", "indent", "separator",
+		  "forecolor", "hilitecolor", "textindicator", "separator",
+		  "inserthorizontalrule", "createlink", "insertimage", "inserttable", "htmlmode", "separator",
+		  "popupeditor", "separator", "showhelp", "about" ]
 		];
 
 	this.fontname = {
@@ -103,6 +113,22 @@ HTMLArea.Config = function () {
 		"Formatted": "pre"
 	};
 
+	this.customSelects = {};
+
+	function cut_copy_paste(e, cmd, obj) {
+		try {
+			e.execCommand(cmd);
+		} catch (e) {
+			if (HTMLArea.is_gecko) {
+				alert("Some revisions of Mozilla/Gecko do not support programatic " +
+				      "access to cut/copy/paste functions, for security reasons.  " +
+				      "Your browser is one of them.  Please use the standard key combinations:\n" +
+				      "CTRL-X for cut, CTRL-C for copy, CTRL-V for paste.");
+				obj.element.style.display = "none";
+			}
+		}
+	};
+
 	// ADDING CUSTOM BUTTONS: please read below!
 	// format of the btnList elements is "ID: [ ToolTip, Icon, Enabled in text mode?, ACTION ]"
 	//    - ID: unique ID for the button.  If the button calls document.execCommand
@@ -115,33 +141,38 @@ HTMLArea.Config = function () {
 	//              These 2 parameters makes it possible for you to use the same
 	//              handler for more HTMLArea objects or for more different buttons.
 	//    - ToolTip: default tooltip, for cases when it is not defined in the -lang- file (HTMLArea.I18N)
-	//    - Icon: path to an icon image file for the button (FIXME: to be changed! use one image for all buttons!)
+	//    - Icon: path to an icon image file for the button (TODO: use one image for all buttons!)
 	//    - Enabled in text mode: if false the button gets disabled for text-only mode; otherwise enabled all the time.
 	this.btnList = {
-		bold: [ "Bold", "ed_format_bold.gif", false, function(e) {e.execCommand("bold");} ],
-		italic: [ "Italic", "ed_format_italic.gif", false, function(e) {e.execCommand("italic");} ],
-		underline: [ "Underline", "ed_format_underline.gif", false, function(e) {e.execCommand("underline");} ],
-		strikethrough: [ "Strikethrough", "ed_format_strike.gif", false, function(e) {e.execCommand("strikethrough");} ],
-		subscript: [ "Subscript", "ed_format_sub.gif", false, function(e) {e.execCommand("subscript");} ],
-		superscript: [ "Superscript", "ed_format_sup.gif", false, function(e) {e.execCommand("superscript");} ],
-		justifyleft: [ "Justify Left", "ed_align_left.gif", false, function(e) {e.execCommand("justifyleft");} ],
-		justifycenter: [ "Justify Center", "ed_align_center.gif", false, function(e) {e.execCommand("justifycenter");} ],
-		justifyright: [ "Justify Right", "ed_align_right.gif", false, function(e) {e.execCommand("justifyright");} ],
-		justifyfull: [ "Justify Full", "ed_align_justify.gif", false, function(e) {e.execCommand("justifyfull");} ],
-		insertorderedlist: [ "Ordered List", "ed_list_num.gif", false, function(e) {e.execCommand("insertorderedlist");} ],
-		insertunorderedlist: [ "Bulleted List", "ed_list_bullet.gif", false, function(e) {e.execCommand("insertunorderedlist");} ],
-		outdent: [ "Decrease Indent", "ed_indent_less.gif", false, function(e) {e.execCommand("outdent");} ],
-		indent: [ "Increase Indent", "ed_indent_more.gif", false, function(e) {e.execCommand("indent");} ],
-		forecolor: [ "Font Color", "ed_color_fg.gif", false, function(e) {e.execCommand("forecolor");} ],
-		hilitecolor: [ "Background Color", "ed_color_bg.gif", false, function(e) {e.execCommand("hilitecolor");} ],
-		inserthorizontalrule: [ "Horizontal Rule", "ed_hr.gif", false, function(e) {e.execCommand("inserthorizontalrule");} ],
-		createlink: [ "Insert Web Link", "ed_link.gif", false, function(e) {e.execCommand("createlink", true);} ],
-		insertimage: [ "Insert Image", "ed_image.gif", false, function(e) {e.execCommand("insertimage");} ],
-		inserttable: [ "Insert Table", "insert_table.gif", false, function(e) {e.execCommand("inserttable");} ],
-		htmlmode: [ "Toggle HTML Source", "ed_html.gif", true, function(e) {e.execCommand("htmlmode");} ],
-		popupeditor: [ "Enlarge Editor", "fullscreen_maximize.gif", true, function(e) {e.execCommand("popupeditor");} ],
-		about: [ "About this editor", "ed_about.gif", true, function(e) {e.execCommand("about");} ],
-		showhelp: [ "Help using editor", "ed_help.gif", true, function(e) {e.execCommand("showhelp");} ]
+		bold: [ "Bold", "images/ed_format_bold.gif", false, function(e) {e.execCommand("bold");} ],
+		italic: [ "Italic", "images/ed_format_italic.gif", false, function(e) {e.execCommand("italic");} ],
+		underline: [ "Underline", "images/ed_format_underline.gif", false, function(e) {e.execCommand("underline");} ],
+		strikethrough: [ "Strikethrough", "images/ed_format_strike.gif", false, function(e) {e.execCommand("strikethrough");} ],
+		subscript: [ "Subscript", "images/ed_format_sub.gif", false, function(e) {e.execCommand("subscript");} ],
+		superscript: [ "Superscript", "images/ed_format_sup.gif", false, function(e) {e.execCommand("superscript");} ],
+		justifyleft: [ "Justify Left", "images/ed_align_left.gif", false, function(e) {e.execCommand("justifyleft");} ],
+		justifycenter: [ "Justify Center", "images/ed_align_center.gif", false, function(e) {e.execCommand("justifycenter");} ],
+		justifyright: [ "Justify Right", "images/ed_align_right.gif", false, function(e) {e.execCommand("justifyright");} ],
+		justifyfull: [ "Justify Full", "images/ed_align_justify.gif", false, function(e) {e.execCommand("justifyfull");} ],
+		insertorderedlist: [ "Ordered List", "images/ed_list_num.gif", false, function(e) {e.execCommand("insertorderedlist");} ],
+		insertunorderedlist: [ "Bulleted List", "images/ed_list_bullet.gif", false, function(e) {e.execCommand("insertunorderedlist");} ],
+		outdent: [ "Decrease Indent", "images/ed_indent_less.gif", false, function(e) {e.execCommand("outdent");} ],
+		indent: [ "Increase Indent", "images/ed_indent_more.gif", false, function(e) {e.execCommand("indent");} ],
+		forecolor: [ "Font Color", "images/ed_color_fg.gif", false, function(e) {e.execCommand("forecolor");} ],
+		hilitecolor: [ "Background Color", "images/ed_color_bg.gif", false, function(e) {e.execCommand("hilitecolor");} ],
+		inserthorizontalrule: [ "Horizontal Rule", "images/ed_hr.gif", false, function(e) {e.execCommand("inserthorizontalrule");} ],
+		createlink: [ "Insert Web Link", "images/ed_link.gif", false, function(e) {e.execCommand("createlink", true);} ],
+		insertimage: [ "Insert Image", "images/ed_image.gif", false, function(e) {e.execCommand("insertimage");} ],
+		inserttable: [ "Insert Table", "images/insert_table.gif", false, function(e) {e.execCommand("inserttable");} ],
+		htmlmode: [ "Toggle HTML Source", "images/ed_html.gif", true, function(e) {e.execCommand("htmlmode");} ],
+		popupeditor: [ "Enlarge Editor", "images/fullscreen_maximize.gif", true, function(e) {e.execCommand("popupeditor");} ],
+		about: [ "About this editor", "images/ed_about.gif", true, function(e) {e.execCommand("about");} ],
+		showhelp: [ "Help using editor", "images/ed_help.gif", true, function(e) {e.execCommand("showhelp");} ],
+		undo: [ "Undoes your last action", "images/ed_undo.gif", false, function(e) {e.execCommand("undo");} ],
+		redo: [ "Redoes your last action", "images/ed_redo.gif", false, function(e) {e.execCommand("redo");} ],
+		cut: [ "Cut selection", "images/ed_cut.gif", false, cut_copy_paste ],
+		copy: [ "Copy selection", "images/ed_copy.gif", false, cut_copy_paste ],
+		paste: [ "Paste from clipboard", "images/ed_paste.gif", false, cut_copy_paste ]
 	};
 	/* ADDING CUSTOM BUTTONS
 	 * ---------------------
@@ -188,15 +219,48 @@ HTMLArea.Config = function () {
  *      textMode : false,            // disabled in text mode
  *      action   : function(editor) { // called when the button is clicked
  *                   editor.surroundHTML('<span class="hilite">', '</span>');
- *                 }
+ *                 },
+ *      context  : "p"               // will be disabled if outside a <p> element
  *    });
  */
-HTMLArea.Config.prototype.registerButton = function(id, tooltip, image, textMode, action) {
-	switch (typeof id) {
-	    case "string": this.btnList[id] = [ tooltip, image, textMode, action ]; break;
-	    case "object": this.btnList[id.id] = [ id.tooltip, id.image, id.textMode, id.action ]; break;
-	    default: alert("HTMLArea.Config::registerButton -- invalid arguments");
+HTMLArea.Config.prototype.registerButton = function(id, tooltip, image, textMode, action, context) {
+	var the_id;
+	if (typeof id == "string") {
+		the_id = id;
+	} else if (typeof id == "object") {
+		the_id = id.id;
+	} else {
+		alert("ERROR [HTMLArea.Config::registerButton]:\ninvalid arguments");
+		return false;
 	}
+	// check for existing id
+	if (typeof this.customSelects[the_id] != "undefined") {
+		alert("WARNING [HTMLArea.Config::registerDropdown]:\nA dropdown with the same ID already exists.");
+	}
+	if (typeof this.btnList[the_id] != "undefined") {
+		alert("WARNING [HTMLArea.Config::registerDropdown]:\nA button with the same ID already exists.");
+	}
+	switch (typeof id) {
+	    case "string": this.btnList[id] = [ tooltip, image, textMode, action, context ]; break;
+	    case "object": this.btnList[id.id] = [ id.tooltip, id.image, id.textMode, id.action, id.context ]; break;
+	}
+};
+
+/** The following helper function registers a dropdown box with the editor
+ * configuration.  You still have to add it to the toolbar, same as with the
+ * buttons.  Call it like this:
+ *
+ * FIXME: add example
+ */
+HTMLArea.Config.prototype.registerDropdown = function(object) {
+	// check for existing id
+	if (typeof this.customSelects[object.id] != "undefined") {
+		alert("WARNING [HTMLArea.Config::registerDropdown]:\nA dropdown with the same ID already exists.");
+	}
+	if (typeof this.btnList[object.id] != "undefined") {
+		alert("WARNING [HTMLArea.Config::registerDropdown]:\nA button with the same ID already exists.");
+	}
+	this.customSelects[object.id] = object;
 };
 
 /** Helper function: replace all TEXTAREA-s in the document with HTMLArea-s. */
@@ -219,9 +283,6 @@ HTMLArea.prototype._createToolbar = function () {
 	this._toolbar = toolbar;
 	toolbar.className = "toolbar";
 	toolbar.unselectable = "1";
-	if (editor.config.debug) {
-		toolbar.style.border = "1px solid red";
-	}
 	var tb_row = null;
 	var tb_objects = new Object();
 	this._toolbarObjects = tb_objects;
@@ -280,6 +341,8 @@ HTMLArea.prototype._createToolbar = function () {
 		var options = null;
 		var el = null;
 		var cmd = null;
+		var customSelects = editor.config.customSelects;
+		var context = null;
 		switch (txt) {
 		    case "fontsize":
 		    case "fontname":
@@ -294,6 +357,17 @@ HTMLArea.prototype._createToolbar = function () {
 			options = editor.config[txt];
 			cmd = txt;
 			break;
+		    default:
+			// try to fetch it from the list of registered selects
+			cmd = txt;
+			var dropdown = customSelects[cmd];
+			if (typeof dropdown != "undefined") {
+				options = dropdown.options;
+				context = dropdown.context;
+			} else {
+				alert("ERROR [createSelect]:\nCan't find the requested dropdown definition");
+			}
+			break;
 		}
 		if (options) {
 			el = document.createElement("select");
@@ -303,7 +377,8 @@ HTMLArea.prototype._createToolbar = function () {
 				enabled : true, // is it enabled?
 				text	: false, // enabled in text mode?
 				cmd	: cmd, // command ID
-				state	: setButtonStatus // for changing state
+				state	: setButtonStatus, // for changing state
+				context : context
 			};
 			tb_objects[txt] = obj;
 			for (var i in options) {
@@ -368,7 +443,8 @@ HTMLArea.prototype._createToolbar = function () {
 				active	: false, // is it pressed?
 				text	: btn[2], // enabled in text mode?
 				cmd	: btn[3], // the command ID
-				state	: setButtonStatus // for changing state
+				state	: setButtonStatus, // for changing state
+				context : btn[4] || null // enabled in a certain context?
 			};
 			tb_objects[txt] = obj;
 			// handlers to emulate nice flat toolbar buttons
@@ -396,12 +472,14 @@ HTMLArea.prototype._createToolbar = function () {
 				if (obj.enabled) with (HTMLArea) {
 					_removeClass(el, "buttonActive");
 					_removeClass(el, "buttonHover");
-					obj.cmd(editor, obj.name);
+					obj.cmd(editor, obj.name, obj);
 					_stopEvent(is_ie ? window.event : ev);
 				}
 			});
 			var img = document.createElement("img");
 			img.src = editor.imgURL(btn[1]);
+			img.style.width = "18px";
+			img.style.height = "18px";
 			el.appendChild(img);
 		} else if (!el) {
 			el = createSelect(txt);
@@ -416,16 +494,53 @@ HTMLArea.prototype._createToolbar = function () {
 		return el;
 	};
 
+	var first = true;
 	for (var i in this.config.toolbar) {
+		if (!first) {
+			createButton("linebreak");
+		} else {
+			first = false;
+		}
 		var group = this.config.toolbar[i];
 		for (var j in group) {
-			createButton(group[j]);
+			var code = group[j];
+			if (/^([IT])\[(.*?)\]/.test(code)) {
+				// special case, create text label
+				var l7ed = RegExp.$1 == "I"; // localized?
+				var label = RegExp.$2;
+				if (l7ed) {
+					label = HTMLArea.I18N.custom[label];
+				}
+				var tb_cell = document.createElement("td");
+				tb_row.appendChild(tb_cell);
+				tb_cell.className = "label";
+				tb_cell.innerHTML = label;
+			} else {
+				createButton(code);
+			}
 		}
 	}
 
 	this._htmlArea.appendChild(toolbar);
 };
 
+HTMLArea.prototype._createStatusBar = function() {
+	var div = document.createElement("div");
+	div.className = "statusBar";
+	this._htmlArea.appendChild(div);
+	this._statusBar = div;
+	div.appendChild(document.createTextNode(HTMLArea.I18N.msg["Path"] + ": "));
+	// creates a holder for the path view
+	div = document.createElement("span");
+	div.className = "statusBarTree";
+	this._statusBarTree = div;
+	this._statusBar.appendChild(div);
+	if (!this.config.statusBar) {
+		// disable it...
+		div.style.display = "none";
+	}
+};
+
 // Creates the HTMLArea object and replaces the textarea with it.
 HTMLArea.prototype.generate = function () {
 	var editor = this;	// we'll need "this" in some nested functions
@@ -472,12 +587,18 @@ HTMLArea.prototype.generate = function () {
 
 	this._iframe = iframe;
 
+	// creates & appends the status bar, if the case
+	this._createStatusBar();
+
 	// remove the default border as it keeps us from computing correctly
 	// the sizes.  (somebody tell me why doesn't this work in IE)
-	// iframe.style.border = "none";
+
+	if (!HTMLArea.is_ie) {
+		iframe.style.borderWidth = "1px";
 	// iframe.frameBorder = "1";
 	// iframe.marginHeight = "0";
 	// iframe.marginWidth = "0";
+	}
 
 	// size the IFRAME according to user's prefs or initial textarea
 	var height = (this.config.height == "auto" ? (this._ta_size.h + "px") : this.config.height);
@@ -485,10 +606,16 @@ HTMLArea.prototype.generate = function () {
 	var width = (this.config.width == "auto" ? (this._ta_size.w + "px") : this.config.width);
 	width = parseInt(width);
 
+	if (!HTMLArea.is_ie) {
+		height -= 2;
+		width -= 2;
+	}
+
 	iframe.style.width = width + "px";
 	if (this.config.sizeIncludesToolbar) {
 		// substract toolbar height
 		height -= this._toolbar.offsetHeight;
+		height -= this._statusBar.offsetHeight;
 	}
 	if (height < 0) {
 		height = 0;
@@ -562,6 +689,9 @@ HTMLArea.prototype.setMode = function(mode) {
 		this._textArea.value = this.getHTML();
 		this._iframe.style.display = "none";
 		this._textArea.style.display = "block";
+		if (this.config.statusBar) {
+			this._statusBar.innerHTML = HTMLArea.I18N.msg["TEXT_MODE"];
+		}
 		break;
 	    case "wysiwyg":
 		if (HTMLArea.is_gecko) {
@@ -575,6 +705,11 @@ HTMLArea.prototype.setMode = function(mode) {
 			// we need to refresh that info for Moz-1.3a
 			this._doc.designMode = "on";
 		}
+		if (this.config.statusBar) {
+			this._statusBar.innerHTML = '';
+			this._statusBar.appendChild(document.createTextNode(HTMLArea.I18N.msg["Path"] + ": "));
+			this._statusBar.appendChild(this._statusBarTree);
+		}
 		break;
 	    default:
 		alert("Mode <" + mode + "> not defined!");
@@ -584,10 +719,42 @@ HTMLArea.prototype.setMode = function(mode) {
 	this.focusEditor();
 };
 
+/***************************************************
+ *  Category: PLUGINS
+ ***************************************************/
+
+// Create the specified plugin and register it with this HTMLArea
+HTMLArea.prototype.registerPlugin = function(pluginName) {
+	this.plugins[pluginName] = eval("new " + pluginName + "(this);");
+};
+
+// static function that loads the required plugin and lang file, based on the
+// language loaded already for HTMLArea.  You better make sure that the plugin
+// _has_ that language, otherwise shit might happen ;-)
+HTMLArea.loadPlugin = function(pluginName) {
+	var editorurl = '';
+	if (typeof _editor_url != "undefined") {
+		editorurl = _editor_url + "/";
+	}
+	var dir = editorurl + "plugins/" + pluginName;
+	var plugin = pluginName.replace(/([a-z])([A-Z])([a-z])/g,
+					function (str, l1, l2, l3) {
+						return l1 + "-" + l2.toLowerCase() + l3;
+					}).toLowerCase() + ".js";
+	document.write("<script type='text/javascript' src='" + dir + "/" + plugin + "'></script>");
+	document.write("<script type='text/javascript' src='" + dir + "/lang/" + HTMLArea.I18N.lang + ".js'></script>");
+};
+
 /***************************************************
  *  Category: EDITOR UTILITIES
  ***************************************************/
 
+HTMLArea.prototype.forceRedraw = function() {
+	this._doc.body.style.visibility = "hidden";
+	this._doc.body.style.visibility = "visible";
+	// this._doc.body.innerHTML = this.getInnerHTML();
+};
+
 // focuses the iframe window.  returns a reference to the editor document.
 HTMLArea.prototype.focusEditor = function() {
 	switch (this._editMode) {
@@ -599,16 +766,100 @@ HTMLArea.prototype.focusEditor = function() {
 };
 
 // updates enabled/disable/active state of the toolbar elements
-HTMLArea.prototype.updateToolbar = function() {
+HTMLArea.prototype.updateToolbar = function(noStatus) {
 	var doc = this._doc;
 	var text = (this._editMode == "textmode");
+	var ancestors = null;
+	if (!text) {
+		ancestors = this.getAllAncestors();
+		if (this.config.statusBar && !noStatus) {
+			this._statusBarTree.innerHTML = ''; // clear
+			for (var i = ancestors.length; --i >= 0;) {
+				var el = ancestors[i];
+				if (!el) {
+					// hell knows why we get here; this
+					// could be a classic example of why
+					// it's good to check for conditions
+					// that are impossible to happen ;-)
+					continue;
+				}
+				var a = document.createElement("a");
+				a.href = "#";
+				a.el = el;
+				a.editor = this;
+				a.onclick = function() {
+					this.blur();
+					this.editor.selectNodeContents(this.el);
+					this.editor.updateToolbar(true);
+					return false;
+				};
+				a.oncontextmenu = function() {
+					// TODO: add context menu here
+					this.blur();
+					var info = "Inline style:\n\n";
+					info += this.el.style.cssText.split(/;\s*/).join(";\n");
+					alert(info);
+					return false;
+				};
+				var txt = el.tagName.toLowerCase();
+				a.title = el.style.cssText;
+				if (el.id) {
+					txt += "#" + el.id;
+				}
+				if (el.className) {
+					txt += "." + el.className;
+				}
+				a.appendChild(document.createTextNode(txt));
+				this._statusBarTree.appendChild(a);
+				if (i != 0) {
+					this._statusBarTree.appendChild(document.createTextNode(String.fromCharCode(0xbb)));
+				}
+			}
+		}
+	}
 	for (var i in this._toolbarObjects) {
 		var btn = this._toolbarObjects[i];
 		var cmd = i;
-		btn.state("enabled", !text || btn.text);
+		var inContext = true;
+		if (btn.context && !text) {
+			inContext = false;
+			var context = btn.context;
+			var attrs = [];
+			if (/(.*)\[(.*?)\]/.test(context)) {
+				context = RegExp.$1;
+				attrs = RegExp.$2.split(",");
+			}
+			context = context.toLowerCase();
+			var match = (context == "*");
+			for (var k in ancestors) {
+				if (!ancestors[k]) {
+					// the impossible really happens.
+					continue;
+				}
+				if (match || (ancestors[k].tagName.toLowerCase() == context)) {
+					inContext = true;
+					for (var ka in attrs) {
+						if (!eval("ancestors[k]." + attrs[ka])) {
+							inContext = false;
+							break;
+						}
+					}
+					if (inContext) {
+						break;
+					}
+				}
+			}
+		}
+		btn.state("enabled", (!text || btn.text) && inContext);
 		if (typeof cmd == "function") {
 			continue;
 		}
+		// look-it-up in the custom dropdown boxes
+		var dropdown = this.config.customSelects[cmd];
+		if ((!text || btn.text) && (typeof dropdown != "undefined")) {
+			dropdown.refresh(this);
+			continue;
+		}
 		switch (cmd) {
 		    case "fontname":
 		    case "fontsize":
@@ -642,6 +893,10 @@ HTMLArea.prototype.updateToolbar = function() {
 				try {with (btn.element.style) {
 					backgroundColor = HTMLArea._makeColor(
 						doc.queryCommandValue(HTMLArea.is_ie ? "backcolor" : "hilitecolor"));
+					if (/transparent/i.test(backgroundColor)) {
+						// Mozilla
+						backgroundColor = HTMLArea._makeColor(doc.queryCommandValue("backcolor"));
+					}
 					color = HTMLArea._makeColor(doc.queryCommandValue("forecolor"));
 					fontFamily = doc.queryCommandValue("fontname");
 					fontWeight = doc.queryCommandState("bold") ? "bold" : "normal";
@@ -672,35 +927,90 @@ HTMLArea.prototype.insertNodeAtSelection = function(toBeInserted) {
 		range.deleteContents();
 		var node = range.startContainer;
 		var pos = range.startOffset;
-		range = this._createRange();
 		switch (node.nodeType) {
 		    case 3: // Node.TEXT_NODE
 			// we have to split it at the caret position.
 			if (toBeInserted.nodeType == 3) {
 				// do optimized insertion
 				node.insertData(pos, toBeInserted.data);
+				range = this._createRange();
 				range.setEnd(node, pos + toBeInserted.length);
 				range.setStart(node, pos + toBeInserted.length);
+				sel.addRange(range);
 			} else {
 				node = node.splitText(pos);
+				var selnode = toBeInserted;
+				if (toBeInserted.nodeType == 11 /* Node.DOCUMENT_FRAGMENT_NODE */) {
+					selnode = selnode.firstChild;
+				}
 				node.parentNode.insertBefore(toBeInserted, node);
-				range.setStart(node, 0);
-				range.setEnd(node, 0);
+				this.selectNodeContents(selnode);
+				this.updateToolbar();
 			}
 			break;
 		    case 1: // Node.ELEMENT_NODE
-			node = node.childNodes[pos];
-			node.parentNode.insertBefore(toBeInserted, node);
-			range.setStart(node, 0);
-			range.setEnd(node, 0);
+			var selnode = toBeInserted;
+			if (toBeInserted.nodeType == 11 /* Node.DOCUMENT_FRAGMENT_NODE */) {
+				selnode = selnode.firstChild;
+			}
+			node.insertBefore(toBeInserted, node.childNodes[pos]);
+			this.selectNodeContents(selnode);
+			this.updateToolbar();
 			break;
 		}
-		sel.addRange(range);
 	} else {
 		return null;	// this function not yet used for IE <FIXME>
 	}
 };
 
+// Returns the deepest node that contains both endpoints of the selection.
+HTMLArea.prototype.getParentElement = function() {
+	var sel = this._getSelection();
+	var range = this._createRange(sel);
+	if (HTMLArea.is_ie) {
+		return range.parentElement ? range.parentElement() : this._doc.body;
+	} else {
+		var p = range.commonAncestorContainer;
+		while (p.nodeType == 3) {
+			p = p.parentNode;
+		}
+		return p;
+	}
+};
+
+// Returns an array with all the ancestor nodes of the selection.
+HTMLArea.prototype.getAllAncestors = function() {
+	var p = this.getParentElement();
+	var a = [];
+	while (p && (p.nodeType == 1) && (p.tagName.toLowerCase() != 'body')) {
+		a.push(p);
+		p = p.parentNode;
+	}
+	a.push(this._doc.body);
+	return a;
+};
+
+// Selects the contents inside the given node
+HTMLArea.prototype.selectNodeContents = function(node, pos) {
+	this.focusEditor();
+	this.forceRedraw();
+	var range;
+	var collapsed = (typeof pos != "undefined");
+	if (HTMLArea.is_ie) {
+		range = this._doc.body.createTextRange();
+		range.moveToElementText(node);
+		(collapsed) && range.collapse(pos);
+		range.select();
+	} else {
+		var sel = this._getSelection();
+		range = this._doc.createRange();
+		range.selectNodeContents(node);
+		(collapsed) && range.collapse(pos);
+		sel.removeAllRanges();
+		sel.addRange(range);
+	}
+};
+
 /** Call this function to insert HTML code at the current position.  It deletes
  * the selection, if any.
  */
@@ -725,7 +1035,7 @@ HTMLArea.prototype.insertHTML = function(html) {
 
 /**
  *  Call this function to surround the existing HTML code in the selection with
- *  your tags.
+ *  your tags.  FIXME: buggy!  This function will be deprecated "soon".
  */
 HTMLArea.prototype.surroundHTML = function(startTag, endTag) {
 	var html = this.getSelectedHTML();
@@ -846,7 +1156,14 @@ HTMLArea.prototype._comboSelected = function(el, txt) {
 		(HTMLArea.is_ie) && (value = "<" + value + ">");
 		this.execCommand(txt, false, value);
 		break;
-	    default: alert("FIXME: combo box " + txt + " not implemented");
+	    default:
+		// try to look it up in the registered dropdowns
+		var dropdown = this.config.customSelects[txt];
+		if (typeof dropdown != "undefined") {
+			dropdown.action(this);
+		} else {
+			alert("FIXME: combo box " + txt + " not implemented");
+		}
 	}
 };
 
@@ -861,7 +1178,9 @@ HTMLArea.prototype.execCommand = function(cmdID, UI, param) {
 		(HTMLArea.is_ie) && (cmdID = "backcolor");
 	    case "forecolor":
 		this._popupDialog("select_color.html", function(color) {
-			editor._doc.execCommand(cmdID, false, "#" + color);
+			if (color) { // selection not canceled
+				editor._doc.execCommand(cmdID, false, "#" + color);
+			}
 		}, HTMLArea._colorToRgb(this._doc.queryCommandValue(cmdID)));
 		break;
 	    case "createlink":
@@ -878,7 +1197,7 @@ HTMLArea.prototype.execCommand = function(cmdID, UI, param) {
 	    case "popupeditor":
 		if (HTMLArea.is_ie) {
 			window.open(this.popupURL("fullscreen.html"), "ha_fullscreen",
-				    "toolbar=no,location=no,directories=no,status=yes,menubar=no," +
+				    "toolbar=no,location=no,directories=no,status=no,menubar=no," +
 				    "scrollbars=no,resizable=yes,width=640,height=480");
 		} else {
 			window.open(this.popupURL("fullscreen.html"), "ha_fullscreen",
@@ -967,8 +1286,12 @@ HTMLArea.prototype._editorEvent = function(ev) {
 	}
 	*/
 	// update the toolbar state after some time
-	setTimeout(function() {
+	if (editor._timerToolbar) {
+		clearTimeout(editor._timerToolbar);
+	}
+	editor._timerToolbar = setTimeout(function() {
 		editor.updateToolbar();
+		editor._timerToolbar = null;
 	}, 50);
 };
 
@@ -1141,7 +1464,7 @@ HTMLArea.isBlockElement = function(el) {
 };
 
 HTMLArea.needsClosingTag = function(el) {
-	var closingTags = " script style div span ";
+	var closingTags = " script style div span tr td tbody table em strong font a ";
 	return (closingTags.indexOf(" " + el.tagName.toLowerCase() + " ") != -1);
 };
 
@@ -1287,8 +1610,12 @@ HTMLArea.prototype._popupDialog = function(url, action, init) {
 
 // paths
 
-HTMLArea.prototype.imgURL = function(file) {
-	return this.config.editorURL + this.config.imgURL + file;
+HTMLArea.prototype.imgURL = function(file, plugin) {
+	if (typeof plugin == "undefined") {
+		return this.config.editorURL + file;
+	} else {
+		return this.config.editorURL + "plugins/" + plugin + "/img/" + file;
+	}
 };
 
 HTMLArea.prototype.popupURL = function(file) {
